==> es7s-3.93.1.dev0.tar.gz: es7s-3.93.1.dev0/es7s/shared/pt_.py <==
# ------------------------------------------------------------------------------
#  es7s/core
#  (c) 2024 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
import typing as t
from dataclasses import dataclass

import pytermor as pt
from pytermor import FT, RT, get_terminal_width, IRenderer, Composite

Size = int | float | None


AUTO: Size = None


@dataclass(frozen=True)
class ElasticSetup:
    min_width: Size = AUTO
    max_width: Size = AUTO
    align: pt.Align | str = pt.Align.LEFT

    @classmethod
    def size_to_chars(cls, size: Size, container_width: int) -> int:
        if size is None:
            return container_width
        if isinstance(size, int):
            return size
        if isinstance(size, float):
            return round(size * container_width)
        raise TypeError(f"Size should be int/float/None, got: {type(size)}")


class ElasticFragment(pt.Fragment, t.Sized):
    def __init__(
        self,
        string: str = "",
        fmt: FT = None,
        *,
        close_this: bool = True,
        close_prev: bool = False,
        es: ElasticSetup = ElasticSetup(),
    ):
        self._elastic_setup = es
        super().__init__(string, fmt, close_this=close_this, close_prev=close_prev)

    @property
    def elastic_setup(self) -> ElasticSetup:
        return self._elastic_setup

    def get_min_width(self, container_width: int) -> int:
        if self._elastic_setup.min_width:
            return ElasticSetup.size_to_chars(self._elastic_setup.min_width, container_width)
        return len(self)

    def get_max_width(self, container_width: int) -> int:
        if self._elastic_setup.max_width:
            return ElasticSetup.size_to_chars(self._elastic_setup.max_width, container_width)
        return len(self)


class ElasticContainer(pt.Composite):
    def __init__(self, *parts: RT, width: Size = AUTO, gap: Size = 2):
        super().__init__(*parts)
        self._width: Size = width
        self._gap: Size = gap

    def __add__(self, other: RT) -> Composite:
        return super().__add__(other)

    def render(self, renderer: IRenderer | t.Type[IRenderer] = None) -> str:
        avail_width = ElasticSetup.size_to_chars(self._width, get_terminal_width())
        if avail_width < 1:
            return ""

        min_widths = []
        max_widths = []
        for part in self._parts:
            if isinstance(part, ElasticContainer):
                raise NotImplementedError

            if isinstance(part, ElasticFragment):
                min_widths.append(part.get_min_width(avail_width))
                max_widths.append(part.get_max_width(avail_width))
            else:
                content_width = len(part)
                min_widths.append(content_width)
                max_widths.append(content_width)

        gap_width = ElasticSetup.size_to_chars(self._gap, avail_width)
        gaps_width = gap_width * (len(self._parts) - 1)
        min_width = sum(min_widths) + gaps_width
        max_width = sum(max_widths) + gaps_width

        if min_width < avail_width < max_width:
            result_widths = min_widths.copy()
            free_width = avail_width - min_width
            idx = 0
            while free_width:
                if result_widths[idx] < max_widths[idx]:
                    free_width -= 1
                    result_widths[idx] += 1
                idx += 1
                if idx >= len(self._parts):
                    idx = 0
        elif avail_width <= min_width:
            result_widths = min_widths
        else:
            result_widths = max_widths

        result = ""
        for (idx, part) in enumerate(self._parts):
            if idx > 0:
                result += pt.pad(gap_width)
            result_width = result_widths.pop(0)
            if isinstance(part, ElasticFragment):
                align = part.elastic_setup.align
            elif isinstance(part, pt.FrozenText):
                align = part._align
            else:
                align = pt.Align.LEFT
            tx = pt.Text(
                *part.as_fragments(),
                width=result_width,
                align=align,
                overflow=["", pt.OVERFLOW_CHAR][len(part) > result_width],
            )
            result += tx.render(renderer)

        return result

    def __len__(self) -> int:
        raise NotImplementedError

    @property
    def has_width(self) -> bool:
        return True

==> es7s-3.94.0.dev0.tar.gz: es7s-3.94.0.dev0/es7s/shared/pt_.py <==
# ------------------------------------------------------------------------------
#  es7s/core
#  (c) 2024 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
import typing as t
from dataclasses import dataclass

import pytermor as pt
from pytermor import FT, RT, get_terminal_width, IRenderer, Composite

Size = int | float | None


AUTO: Size = None


@dataclass(frozen=True)
class ElasticSetup:
    min_width: Size = AUTO
    max_width: Size = AUTO
    align: pt.Align | str = pt.Align.LEFT
    overflow: bool = None

    @classmethod
    def size_to_chars(cls, size: Size, container_width: int) -> int:
        if size is None:
            return container_width
        if isinstance(size, int):
            return size
        if isinstance(size, float):
            return round(size * container_width)
        raise TypeError(f"Size should be int/float/None, got: {type(size)}")


class ElasticFragment(pt.Fragment, t.Sized):
    def __init__(
        self,
        string: str = "",
        fmt: FT = None,
        *,
        close_this: bool = True,
        close_prev: bool = False,
        es: ElasticSetup = ElasticSetup(),
    ):
        self._elastic_setup = es
        super().__init__(string, fmt, close_this=close_this, close_prev=close_prev)

    @property
    def elastic_setup(self) -> ElasticSetup:
        return self._elastic_setup

    def get_min_width(self, container_width: int) -> int:
        if self._elastic_setup.min_width:
            return ElasticSetup.size_to_chars(self._elastic_setup.min_width, container_width)
        return len(self)

    def get_max_width(self, container_width: int) -> int:
        if self._elastic_setup.max_width:
            return ElasticSetup.size_to_chars(self._elastic_setup.max_width, container_width)
        return len(self)


class ElasticContainer(pt.Composite):
    def __init__(self, *parts: RT, width: Size = AUTO, gap: Size = 2):
        super().__init__(*parts)
        self._width: Size = width
        self._gap: Size = gap

    def __add__(self, other: RT) -> Composite:
        return super().__add__(other)

    def render(self, renderer: IRenderer | t.Type[IRenderer] = None) -> str:
        avail_width = ElasticSetup.size_to_chars(self._width, get_terminal_width())
        if avail_width < 1:
            return ""

        min_widths = []
        max_widths = []
        for part in self._parts:
            if isinstance(part, ElasticContainer):
                raise NotImplementedError

            if isinstance(part, ElasticFragment):
                min_widths.append(part.get_min_width(avail_width))
                max_widths.append(part.get_max_width(avail_width))
            else:
                content_width = len(part)
                min_widths.append(content_width)
                max_widths.append(content_width)

        gap_width = ElasticSetup.size_to_chars(self._gap, avail_width)
        gaps_width = gap_width * (len(self._parts) - 1)
        min_width = sum(min_widths) + gaps_width
        max_width = sum(max_widths) + gaps_width

        if min_width < avail_width < max_width:
            result_widths = min_widths.copy()
            free_width = avail_width - min_width
            idx = 0
            while free_width:
                if result_widths[idx] < max_widths[idx]:
                    free_width -= 1
                    result_widths[idx] += 1
                idx += 1
                if idx >= len(self._parts):
                    idx = 0
        elif avail_width <= min_width:
            result_widths = min_widths
        else:
            result_widths = max_widths

        result = ""
        for (idx, part) in enumerate(self._parts):
            if idx > 0:
                result += pt.pad(gap_width)

            result_width = result_widths.pop(0)
            use_overflow = len(part) > result_width

            if isinstance(part, ElasticFragment):
                align = part.elastic_setup.align
                if part.elastic_setup.overflow is not None:
                    use_overflow = part.elastic_setup.overflow
            elif isinstance(part, pt.FrozenText):
                align = part._align
            else:
                align = pt.Align.LEFT
            tx = pt.Text(
                *part.as_fragments(),
                width=result_width,
                align=align,
                overflow=["", pt.OVERFLOW_CHAR][use_overflow],
            )
            result += tx.render(renderer)

        return result

    def __len__(self) -> int:
        raise NotImplementedError

    @property
    def has_width(self) -> bool:
        return True


def guess_width(s: str) -> int:
    return sum(pt.guess_char_width(c) for c in s)
